#include "Draw.h"
#include "Tree.h"
#include "Constants.h"
#include <OpenGL/gl.h>
#include "GL/glui.h"
#include <stdio.h>

#include <math.h>
#define PI 3.14159265


//Need to know render mode
extern int renderMode;

//Prototypes for non-public functions
void draw_stick(Tree *tree, Tree *parent);
void draw_trunk(Tree *tree, Tree *parent);
void draw_sphere(Tree *tree, Tree *parent);
void precompute();

//Constants and pre-computations for drawing trunks
#define NUM_STEPS 10
#define DELTA_ANG (2*PI/NUM_STEPS)
double sines[NUM_STEPS+1];
double cosines[NUM_STEPS+1];
const double RAD_TO_DEG = 180/PI;
static int precomputed = 0;

//Display list
#define DL_TREE 1
static int compiled = 0;



void decompile() {
	compiled = 0;
}

/**
 * Draw a tree, according to the right mode.
 * This is the only public function
 */
void draw_tree(Tree *tree) {
	if( compiled ) {
		glCallList(DL_TREE);
	} else {
		glNewList(DL_TREE, GL_COMPILE_AND_EXECUTE);
		
		glColor3f(0.0f, 1.0f, 0.0f);
		if( renderMode == MODE_STICK ) {
			glBegin(GL_LINES);
			draw_stick(tree, NULL);
			glEnd();
		} else if( renderMode == MODE_SPHERE ) {
			draw_sphere(tree, NULL);
		} else {
			if( renderMode == MODE_WIRE )
				 glPolygonMode(GL_FRONT,GL_LINE);
			else
				glPolygonMode(GL_FRONT,GL_FILL);
			
			if( !precomputed )
				precompute();
			
			draw_trunk(tree, NULL);
		}
		
		glEndList();
		compiled = 1;
	}
}


/**
 * Recursive function to draw each segment as a line
 */
void draw_stick(Tree *tree, Tree *parent) {
	int i;
	
	if( parent != NULL && tree->bud == 0 ) {
		glVertex3f(parent->x, parent->y, parent->z);
		glVertex3f(tree->x, tree->y, tree->z);
	}
	
	
	for(i = 0; i < tree->numChildren; i++) {
		draw_stick(tree->children[i], tree);
	}
}


/**
 * Recursive function to draw each segment as a tube.
 * 
 * Need to tilt segments around an axis, since not all branches are straight up.
 * We will use spherical coordinates for this: theta and phi.
 *   - Theta is rotation in ZX plane (rotation around Y axis)
 *   - Phi is rotation down from Y axis (rotate around Z axis, after theta)
 * Note: dist is the length of a segment, but is not the hypotenus of any trianlge
 * we're using to find angles.  So use dx, dy, dz and tangents.
 */
void draw_trunk(Tree *tree, Tree *parent) {
	int i;
	double theta, phi;
	double dist;
	double x0, z0, x1, z1, x2, z2, x3, z3;
	
	
	if( parent != NULL && tree->bud == 0 ) {
		double dx, dy, dz;
		dx = tree->x - parent->x;
		dy = tree->y - parent->y;
		dz = tree->z - parent->z;
		dist = sqrt(dx*dx + dy*dy + dz*dz);
		
		theta = dx == 0? 0 : atan2(-dz, dx);
		phi = dy == 0? 0 : atan2(dx, dy);
		//if( theta < -(PI/2) )	//Uhm, I can't figure this out, but this hack makes my demo tree look right
		//	phi = -phi;
		
		
		//float nd = sqrt(dx*dx + dy*dy);
		//float nx = nd*cos(phi);
		//float ny = nd*sin(phi);
		float nx = -dx;
		float ny = dy;
		if( phi < 0 ) {	//only do positive, the theta rotation will account for negative
			phi = -phi;
			nx = -nx;
		}
		
		glPushMatrix();
		glTranslatef(parent->x, parent->y, parent->z);	//base of branch becomes origin
		
		/*if( dx == -20 ) {
			glBegin(GL_LINES);
			glColor3f(1,1,0);
			glVertex3f(0,0,0);
			glVertex3f(100,0,0);
			glVertex3f(0,0,0);
			glVertex3f(0,100,0);
			glVertex3f(0,0,0);
			glVertex3f(0,0,100);
			glEnd();
		}*/
		
		glRotated(-phi * RAD_TO_DEG, 0, 0, 1);
		
		/*if( dx == -20 ) {
			glBegin(GL_LINES);
			glColor3f(1,0,0);
			glVertex3f(0,0,0);
			glVertex3f(100,0,0);
			glVertex3f(0,0,0);
			glVertex3f(0,100,0);
			glVertex3f(0,0,0);
			glVertex3f(0,0,100);
			glEnd();
		}*/
		
		/*if( dx == -20 ) {
			glBegin(GL_LINES);
			
			float m = fabs(dx);
			if( fabs(dy) > m ) m = fabs(dy);
			if( fabs(dz) > m ) m = fabs(dz);
			
			glColor3f(fabs(dx)/m,fabs(dy)/m,fabs(dz)/m);
			glVertex3f(0,0,0);
			glVertex3f((nx*100),(ny*100),0);
			glEnd();
		printf("dx=%f, dy=%f, dz=%f, theta=%4.3f, phi=%4.3f, nx=%4.3f ny=%4.3f\n",
			   dx, dy, dz, theta*RAD_TO_DEG, phi*RAD_TO_DEG, nx, ny);
			//glColor3f(0,1,0);
		}//*/
		
		glRotated(theta * RAD_TO_DEG, nx, ny, 0);
		
		/*if( dx == -20 ) {
			glBegin(GL_LINES);
			glColor3f(0,0,1);
			glVertex3f(0,0,0);
			glVertex3f(100,0,0);
			glVertex3f(0,0,0);
			glVertex3f(0,100,0);
			glVertex3f(0,0,0);
			glVertex3f(0,0,100);
			glEnd();
			
			glColor3f(0,1,0);
		}
		
		printf("(%4.2f,%4.2f,%4.2f) -> (%4.2f,%4.2f,%4.2f) d=(%4.2f,%4.2f,%4.2f) theta=%4.2f phi=%4.2f\n",
			   parent->x, parent->y, parent->z, 
			   tree->x, tree->y, tree->z, 
			   dx, dy, dz,
			   theta * RAD_TO_DEG,
			   -phi * RAD_TO_DEG
			   );
		*/
		
		glBegin(GL_QUADS);
		for(i = 0; i < NUM_STEPS; i++) {
			//circle is in XZ plane (treat Z like Y in normal 2D)
			//start at ang=0, move ccw. Build rects CW relative to it's face.
			
			x0 = (parent->thickness/2) * cosines[i];
			z0 = (parent->thickness/2) * sines[i];
			
			x1 = (tree->thickness/2) * cosines[i];
			z1 = (tree->thickness/2) * sines[i];
			
			x2 = (tree->thickness/2) * cosines[i+1];
			z2 = (tree->thickness/2) * sines[i+1];
			
			x3 = (parent->thickness/2) * cosines[i+1];
			z3 = (parent->thickness/2) * sines[i+1];
			
			glVertex3f(x0, 0, z0);
			glVertex3f(x1, dist, z1);
			glVertex3f(x2, dist, z2);
			glVertex3f(x3, 0, z3);
		}
		glEnd();
		
		glPopMatrix();
	}
	
	
	for(i = 0; i < tree->numChildren; i++) {
		draw_trunk(tree->children[i], tree);
	}
}

void precompute() {
	int i;
	double ang;
	
	for(i = 0; i < NUM_STEPS + 1; i++) {
		ang = i*DELTA_ANG;
		sines[i] = sin(ang);
		cosines[i] = cos(ang);
	}
	
	precomputed = 1;
}

void draw_sphere(Tree *tree, Tree *parent) {
	int i;
	
	if( parent != NULL && tree->bud == 0 ) {
		glBegin(GL_LINES);
		glVertex3f(parent->x, parent->y, parent->z);
		glVertex3f(tree->x, tree->y, tree->z);
		glEnd();
	}
	
	if( tree->thickness != 0 ) {
		glPushMatrix();
		glTranslatef(tree->x, tree->y, tree->z);
		glutSolidSphere(tree->thickness, NUM_STEPS, NUM_STEPS);
		glPopMatrix();
	}
	
	for(i = 0; i < tree->numChildren; i++) {
		draw_sphere(tree->children[i], tree);
	}
}




/**
 * Recursive function to draw each segment as a tube.
 * 
 * Need to tilt segments around an axis, since not all branches are straight up.
 * We will use spherical coordinates for this: theta and phi.
 *   - Theta is rotation in ZX plane (rotation around Y axis)
 *   - Phi is rotation down from Y axis (rotate around Z axis, after theta)
 * Note: dist is the length of a segment, but is not the hypotenus of any trianlge
 * we're using to find angles.  So use dx, dy, dz and tangents.
 * /
void draw_trunk(Tree *tree, Tree *parent) {
	int i;
	double theta, phi;
	double dist;
	double x0, z0, x1, z1, x2, z2, x3, z3;
	
	
	if( parent != NULL ) {
		double dx, dy, dz;
		dx = tree->x - parent->x;
		dy = tree->y - parent->y;
		dz = tree->z - parent->z;
		dist = sqrt(dx*dx + dy*dy + dz*dz);
		
		theta = dx == 0? 0 : atan2(-dz, dx);
		phi = dy == 0? 0 : atan2(dx, dy);
		if( theta < -(PI/2) )	//Uhm, I can't figure this out, but this hack makes my demo tree look right
			phi = -phi;
		
		glPushMatrix();
		glTranslatef(parent->x, parent->y, parent->z);	//base of branch becomes origin
		
		/*if( dx == -20 ) {
		 glBegin(GL_LINES);
		 glColor3f(1,1,0);
		 glVertex3f(0,0,0);
		 glVertex3f(100,0,0);
		 glVertex3f(0,0,0);
		 glVertex3f(0,100,0);
		 glVertex3f(0,0,0);
		 glVertex3f(0,0,100);
		 glEnd();
		 }* /
		
		glRotated(theta * RAD_TO_DEG, 0, 1, 0);
		
		/*if( dx == -20 ) {
		 glBegin(GL_LINES);
		 glColor3f(1,0,0);
		 glVertex3f(0,0,0);
		 glVertex3f(100,0,0);
		 glVertex3f(0,0,0);
		 glVertex3f(0,100,0);
		 glVertex3f(0,0,0);
		 glVertex3f(0,0,100);
		 glEnd();
		 }* /
		
		glRotated(-phi * RAD_TO_DEG, 0, 0, 1);
		
		/*if( dx == -20 ) {
		 glBegin(GL_LINES);
		 glColor3f(0,0,1);
		 glVertex3f(0,0,0);
		 glVertex3f(100,0,0);
		 glVertex3f(0,0,0);
		 glVertex3f(0,100,0);
		 glVertex3f(0,0,0);
		 glVertex3f(0,0,100);
		 glEnd();
		 
		 glColor3f(0,1,0);
		 }
		 
		 printf("(%4.2f,%4.2f,%4.2f) -> (%4.2f,%4.2f,%4.2f) d=(%4.2f,%4.2f,%4.2f) theta=%4.2f phi=%4.2f\n",
		 parent->x, parent->y, parent->z, 
		 tree->x, tree->y, tree->z, 
		 dx, dy, dz,
		 theta * RAD_TO_DEG,
		 -phi * RAD_TO_DEG
		 );
		 * /
		
		glBegin(GL_QUADS);
		for(i = 0; i < NUM_STEPS; i++) {
			//circle is in XZ plane (treat Z like Y in normal 2D)
			//start at ang=0, move ccw. Build rects CW relative to it's face.
			
			x0 = (parent->thickness/2) * cosines[i];
			z0 = (parent->thickness/2) * sines[i];
			
			x1 = (tree->thickness/2) * cosines[i];
			z1 = (tree->thickness/2) * sines[i];
			
			x2 = (tree->thickness/2) * cosines[i+1];
			z2 = (tree->thickness/2) * sines[i+1];
			
			x3 = (parent->thickness/2) * cosines[i+1];
			z3 = (parent->thickness/2) * sines[i+1];
			
			glVertex3f(x0, 0, z0);
			glVertex3f(x1, dist, z1);
			glVertex3f(x2, dist, z2);
			glVertex3f(x3, 0, z3);
		}
		glEnd();
		
		glPopMatrix();
	}
	
	
	for(i = 0; i < tree->numChildren; i++) {
		draw_trunk(tree->children[i], tree);
	}
}//*/