#include "DrawVoxels.h"
//#include "Constants.h"
#include <OpenGL/gl.h>
//#include "GL/glui.h"
#include <stdio.h>

#include <math.h>


//Prototypes for non-public functions
void draw_voxel_grid(VoxelGrid *grid, float size);

//Display list
#define DL_VOXEL 2
static int compiled = 0;
static int compiledSize = 0;
static int compiledShowGrid = 1;
extern int showGrid;


void decompile_voxels() {
	compiled = 0;
}

/**
 * Public function to handle the display of a voxel grid.
 * This will compile a display list the first time, and use it on
 * subsequent calls unless decompile is called, or the size given
 * is different.
 * 
 * @param grid The grid to show
 * @param size Dimension (width/height/depth) of one voxel.
 */
void draw_voxels(VoxelGrid *grid, float size) {
	if( compiled && size == compiledSize && showGrid == compiledShowGrid ) {
		glCallList(DL_VOXEL);
	} else {
		glNewList(DL_VOXEL, GL_COMPILE_AND_EXECUTE);
		
		draw_voxel_grid(grid, size);
		
		glEndList();
		compiled = 1;
		compiledSize = size;
		compiledShowGrid = showGrid;
	}
}

/**
 * Helper method to draw a cube centered at some location with a given
 * size.
 */
void draw_cube(float centerX, float centerY, float centerZ, float size) {
	const float size2 = size/2;
	const float x0 = centerX - size2;
	const float x1 = centerX + size2;
	const float y0 = centerY - size2;
	const float y1 = centerY + size2;
	const float z0 = centerZ - size2;
	const float z1 = centerZ + size2;
	
	glBegin(GL_QUADS);
	
	//X-min face
	glVertex3f(x0, y0, z0);
	glVertex3f(x0, y1, z0);
	glVertex3f(x0, y1, z1);
	glVertex3f(x0, y0, z1);
	
	//X-max face
	glVertex3f(x1, y0, z1);
	glVertex3f(x1, y1, z1);
	glVertex3f(x1, y1, z0);
	glVertex3f(x1, y0, z0);
	
	//Y-min face
	glVertex3f(x1, y0, z0);
	glVertex3f(x0, y0, z0);
	glVertex3f(x0, y0, z1);
	glVertex3f(x1, y0, z1);
	
	//Y-max face
	glVertex3f(x1, y1, z1);
	glVertex3f(x0, y1, z1);
	glVertex3f(x0, y1, z0);
	glVertex3f(x1, y1, z0);
	
	//Z-min face
	glVertex3f(x1, y0, z0);
	glVertex3f(x1, y1, z0);
	glVertex3f(x0, y1, z0);
	glVertex3f(x0, y0, z0);
	
	//Z-max face
	glVertex3f(x0, y0, z1);
	glVertex3f(x0, y1, z1);
	glVertex3f(x1, y1, z1);
	glVertex3f(x1, y0, z1);
	
	glEnd();
}

/**
 * Does the actual drawing.
 */
void draw_voxel_grid(VoxelGrid *grid, float size) {
	printf("drawing voxel grid - size=%5.3f\n",size);
	//Starting x,y,z for drawing
	const float sx = grid->getWidth()/2 * -size - size/2;
	const float sy = 0;
	const float sz = grid->getDepth()/2 * -size - size/2;
	const float size2 = size/2;
	
	//Draw frames first
	if( showGrid ) {
		glColor3f(1.0, 1.0, 0.0);
		glBegin(GL_LINES);
		for(int z = 0; z < grid->getDepth(); z++) {
			for(int y = 0; y < grid->getHeight(); y++) {
				for(int x = 0; x < grid->getWidth(); x++) {
					//printf("Drawing box for (%d, %d, %d)\n", x, y, z);
					//Everyone draws 3 lines - max X, Y, Z. (min-max in one, max in other two)
					//Those on a min-side edge, draw 2 more - min in that dim, max/min-max in other two
					//Those on a double min-side edge draw 1 more - min in those two dims, min-max in other
					const float xoff0 = x * size;
					const float xoff1 = (x+1) * size;
					const float yoff0 = y * size;
					const float yoff1 = (y+1) * size;
					const float zoff0 = z * size;
					const float zoff1 = (z+1) * size;
					
					//Max XY
					glVertex3f(sx + xoff1, sy + yoff1, sz + zoff0);
					glVertex3f(sx + xoff1, sy + yoff1, sz + zoff1);
					//Max XZ
					glVertex3f(sx + xoff1, sy + yoff0, sz + zoff1);
					glVertex3f(sx + xoff1, sy + yoff1, sz + zoff1);
					//Max YZ
					glVertex3f(sx + xoff0, sy + yoff1, sz + zoff1);
					glVertex3f(sx + xoff1, sy + yoff1, sz + zoff1);
					
					//Check one-dim mins
					if( x == 0 ) {
						//Min-Max Y
						glVertex3f(sx + xoff0, sy + yoff0, sz + zoff1);
						glVertex3f(sx + xoff0, sy + yoff1, sz + zoff1);
						//Min-Max Z
						glVertex3f(sx + xoff0, sy + yoff1, sz + zoff0);
						glVertex3f(sx + xoff0, sy + yoff1, sz + zoff1);
					}
					if( y == 0 ) {
						//Min-Max X
						glVertex3f(sx + xoff0, sy + yoff0, sz + zoff1);
						glVertex3f(sx + xoff1, sy + yoff0, sz + zoff1);
						//Min-Max Z
						glVertex3f(sx + xoff1, sy + yoff0, sz + zoff0);
						glVertex3f(sx + xoff1, sy + yoff0, sz + zoff1);
					}
					if( z == 0 ) {
						//Min-Max X
						glVertex3f(sx + xoff0, sy + yoff1, sz + zoff0);
						glVertex3f(sx + xoff1, sy + yoff1, sz + zoff0);
						//Min-Max Y
						glVertex3f(sx + xoff1, sy + yoff0, sz + zoff0);
						glVertex3f(sx + xoff1, sy + yoff1, sz + zoff0);
					}
					
					//Check two-dim mins
					if( y == 0 && z == 0 ) {
						//Min-Max X
						glVertex3f(sx + xoff0, sy + yoff0, sz + zoff0);
						glVertex3f(sx + xoff1, sy + yoff0, sz + zoff0);
					}
					if( x == 0 && z == 0 ) {
						//Min-Max Y
						glVertex3f(sx + xoff0, sy + yoff0, sz + zoff0);
						glVertex3f(sx + xoff0, sy + yoff1, sz + zoff0);
					}
					if( x == 0 && y == 0 ) {
						//Min-Max Z
						glVertex3f(sx + xoff0, sy + yoff0, sz + zoff0);
						glVertex3f(sx + xoff0, sy + yoff0, sz + zoff1);
					}
				}
			}
		}
		glEnd();
	}
	
	//Now draw shadow/shelter indicators
	glColor3f(0.5, 0, 0.5);
	for(int z = 0; z < grid->getDepth(); z++) {
		for(int y = 0; y < grid->getHeight(); y++) {
			for(int x = 0; x < grid->getWidth(); x++) {
				const float xoff = x * size + size2;
				const float yoff = y * size + size2;
				const float zoff = z * size + size2;
				const float shadow = (grid->get(x, y, z) / lightMax) * size;
				if( shadow > size ) {
					//Limit to size of cell, but mark in different color
					glColor3f(0.1, 0, 0.5);
					draw_cube(sx + xoff, sy + yoff, sz + zoff, size);
					glColor3f(0.5, 0, 0.5);
				} else {
					draw_cube(sx + xoff, sy + yoff, sz + zoff, shadow);
				}
			}
		}
	}//*/
}




