/**
 *	Tree Generation with Wind Effects
 *	@author Everett Morse
 *	
 *	This is made for Brigham Young University's CS 658 course "Computer Generated 
 *	Natural Phenomenon" (graduate level).
 *	
 *	Description:
 *	Tree growth has been modeled using space and light competition from voxel-
 *	based shadow models.  They ignore the effects of wind.  This extends the 
 *	voxel-based shadow model to represent the amount of wind reaching each bud
 *	and affect the growth direction.  I also simulate random gusts that may
 *	break off older branches.
 *	
 *	Organization:
 *	This file contains the driver that handles the GUI and main OpenGL display 
 *	functions and launches everything.
 *
 *	Note: Some code is copied from other projects, including the Moon Phases
 *	one which I didn't originally write.  Bits also come from various internet
 *	sources.
 */

/* Standard Includes */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#define PI 3.14159265

/* Open GL, GLUT, GLUi */
#include "GL/glui.h"

/* Project Includes */
#include "Tree.h"
#include "Draw.h"
#include "Writer.h"
#include "Treegen.h"
#include "RandTree.h"
#include "ExampleTree.h"
#include "LightTree.h"
#include "WindTree.h"


/* Settings - largely borrowed from MoonPhase */
#include "Constants.h"
#define FOVY 65.0
double zFar = 1500;
double zNear = 1;
float LightPartialAmbient[] = { 0.3f, 0.3f, 0.3f, 1.0f };
float LightDiffuse[] = { 1.0f, 1.0f, 1.0f, 1.0f };
float LightSpecular[] ={ 0.5f, 0.5f, 0.5f, 1.0f };
float LightPosition[] = { 500.0f, 1000.0f, 1000.0f, 1.0f };


/* Display */
float rotation[16] = {	1.0, 0.0, 0.0, 0.0,
						0.0, 1.0, 0.0, 0.0, 
						0.0, 0.0, 1.0, 0.0,
						0.0, 0.0, 0.0, 1.0 };
int renderMode = MODE_FILL;
int showInfo = 1;
float zoom = 1.0;
float trans[3] = {0.0, 0.0, 0.0};

/* Tree */
Treegen *gen = NULL;
Tree *tree = NULL;

/* Configurable Parameters */
extern float shadowBase;
extern float shadowFactor;
extern float lightMax;
extern int qMax;
lighttree_mode_t ltMode = APICAL;
int showGrid = 1;	//when drawing the voxel grid, show grid or just shadow boxes
extern double WIND_FACTOR;

/* GLUI interface */
int windowId;			//Main Window Id
GLUI *glui;
GLUI_Rotation *rotator;
int smoothShade = 1;
int cull = 1;
char filename[sizeof(GLUI_String)];
//Button Ids
#define B_QUIT 1
#define B_ROTATION 2
#define B_RESET_VIEW 3
#define B_MODE 4
#define B_SHADING 5
#define B_EXAMPLE 6
#define B_RANDOM 7
#define B_STEP 8
#define B_GENERATE 9
#define B_PARAM 10
#define B_THICKNESS 11
#define B_SHOW_INFO 12
#define B_LIGHT 13
#define B_CULL 14
#define B_LOAD 15
#define B_SAVE 16
#define B_ZOOM_IN 17
#define B_ZOOM_OUT 18
#define B_TRANSLATE 19
#define B_WIND 20


/**
 * Set up the camera position and perspective. (Called by draw scene.)
 */
void ResetPerspective(){
	glMatrixMode(GL_PROJECTION);	
	glLoadIdentity();									
	gluPerspective(FOVY, (float)WIDTH/(float)HEIGHT, zNear, zFar);
	
	glMatrixMode(GL_MODELVIEW);							
	glLoadIdentity();	
	gluLookAt(0, 0, 700,		//Eye -- looking down +Z axis towards -Z
			  0, 0, 0,			//Center -- (base of tree translated down a bit)
			  0, 1, 0);			//Up -- +Y is up

	//glViewport(0, 0, WIDTH, HEIGHT);
	GLUI_Master.auto_set_viewport();
}

/**
 * Set up lighting. (Called by draw scene.)
 */
void InitLighting(void){
	
glLightfv(GL_LIGHT0, GL_AMBIENT, LightPartialAmbient);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, LightDiffuse);
	//glLightfv(GL_LIGHT0, GL_SPECULAR, LightSpecular);
	
	//glMaterialfv(GL_FRONT, GL_SPECULAR, LightSpecular);
	//glMateriali(GL_FRONT, GL_SHININESS, 128);
	
	glLightfv( GL_LIGHT0, GL_POSITION, LightPosition );
	
	glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
	glEnable(GL_COLOR_MATERIAL);
	
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
}

/**
 * Set up OpenGL (called by main)
 */
void InitGL(void) {
	ResetPerspective();
	
	if( smoothShade )
		glShadeModel(GL_SMOOTH);
	else
		glShadeModel(GL_FLAT);
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	if( cull ) {
		glEnable(GL_DEPTH_TEST);
		glEnable(GL_CULL_FACE);
		glDepthMask(GL_TRUE);
	}
	glCullFace(GL_BACK);
	
	InitLighting();
	
	//glEnable(GL_BLEND);
	//glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	
	//glDepthFunc(GL_LEQUAL);								
	//glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);	
	
	/*
	if(wireframe)
		glPolygonMode(GL_FRONT,GL_LINE);
	else
		glPolygonMode(GL_FRONT,GL_FILL);
	//*/
}

/**
 * Called by GLUT to draw the window contents (other than GLUi sub-window)
 */
void DrawScene(void) {
	
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);	
	ResetPerspective();
	InitLighting();
	
	//Rotate and translate, then draw
	glPushMatrix();
	glScalef(zoom, zoom, zoom);
	glTranslatef(0, -100, 0);
	glMultMatrixf(rotation);
	glTranslatef(trans[0], trans[1], trans[2]);
	
	//Drawing tree
	if( tree != NULL )
		draw_tree(tree);
	if( gen != NULL && showInfo )
		gen->drawInfo();
	
	glPopMatrix();
	
	glutSwapBuffers();
}

/**
 * Call when the terrain or view angle has changed
 */
void Repaint() {
	glutSetWindow(windowId);
	glutPostRedisplay();
}

/**
 * GLUi has a bug that puts non-visible characters in the text fields
 * it displays.  (Maybe a tab, arrow key press, or something).
 * This function removes those characters.  I limit to only a-zA-Z0-9
 * and one dot character.  All other chars before dot get turned to
 * underscore, all after the dot end the filename.
 */
void sanatizeFilename() {
	int i, j;
	
	i = 0;
	j = -1;
	while( filename[i] ) {
		if( j != -1 ) {
			//After dot
			if( filename[i] == '.' ) {
				filename[j] = '_';	//only allow one "."
				j = i;
			} else if( !((filename[i] >= '0' && filename[i] <= '9') ||
						 (filename[i] >= 'a' && filename[i] <= 'z') ||
						 (filename[i] >= 'A' && filename[i] <= 'Z') )  ) {
				//Truncate once we get garbage
				filename[i] = '\0';
				break;
			}
		} else {
			//Before .
			if( filename[i] == '.' ) {
				j = i;
			} else if( !((filename[i] >= '0' && filename[i] <= '9') ||
						 (filename[i] >= 'a' && filename[i] <= 'z') ||
						 (filename[i] >= 'A' && filename[i] <= 'Z') )  ) {
				//Replace garbage
				filename[i] = '_';
			}
		}
		i++;
	}
	
	glui->sync_live();
}


/**
 * Helper function to handle replacing the current tree with a newly
 * generated one. (Has to delete the old tree, refresh screen, etc.)
 */
void generateTree(Treegen * ngen) {
	if( gen != NULL )
		delete gen;
	if( tree != NULL )
		delete tree;
	
	gen = ngen;
	tree = gen->getTree();
	
	decompile();
	Repaint();
}

/**
 * Called by GLUi when a button has been pressed.
 */
void ControlButton(int id) {
	int c;
	
	switch(id) {
		case B_QUIT:
			exit(0);
			break;
		case B_ROTATION:
			Repaint();
			break;
		case B_RESET_VIEW:
			rotator->reset();
			trans[0] = 0;
			trans[1] = 0;
			trans[2] = 0;
			zoom = 1.0;
			glui->sync_live();
			Repaint();
			break;
		case B_SHADING:
			if( smoothShade )
				glShadeModel(GL_SMOOTH);
			else
				glShadeModel(GL_FLAT);
			Repaint();
			break;
		case B_CULL:
			if( cull ) {
				glEnable(GL_DEPTH_TEST);
				glEnable(GL_CULL_FACE);
				glDepthMask(GL_TRUE);
			} else {
				glDisable(GL_DEPTH_TEST);
				glDisable(GL_CULL_FACE);
			}
			Repaint();
			break;
		case B_MODE:
			decompile();
			Repaint();
			break;
		case B_LOAD:
			
			//Clear out old tree
			if( tree != NULL )
				delete tree;
			
			sanatizeFilename();
			tree = loadFile(filename);
			
			//glui->sync_live();
			decompile();
			Repaint();
			break;
		case B_SAVE:
			
			sanatizeFilename();
			saveFile(filename, tree);
			
			break;
		case B_EXAMPLE:
			generateTree(new ExampleTree());
			break;
		case B_RANDOM:
			generateTree( new RandTree() );
			break;
		case B_LIGHT:
			generateTree( new LightTree() );
			break;
		case B_WIND:
			generateTree( new WindTree() );
			break;
		case B_STEP:
			if( gen != NULL ) {
				gen->step();
				
				decompile();
				Repaint();
			}
			break;
		case B_GENERATE:
			if( gen != NULL ) {
				gen->generate();
				decompile();
				Repaint();
			}
			break;
		case B_THICKNESS:
			if( gen != NULL ) {
				gen->clearThickness();
				gen->calcThickness(thicknessInit);
				
				decompile();
				Repaint();
			}
			break;
		case B_SHOW_INFO:
		case B_TRANSLATE:
			Repaint();
			break;
		case B_ZOOM_IN:
			if( zoom < 10 )
				zoom += 0.5;
			Repaint();
			break;
		case B_ZOOM_OUT:
			if( zoom > 0.5 )
				zoom -= 0.5;
			Repaint();
			break;
		default:
			break;
	}
}

void InitGUI(int windowId) {
	GLUI_Listbox *box;
	GLUI_Spinner *spin;
	
	glui = GLUI_Master.create_glui_subwindow(windowId, GLUI_SUBWINDOW_LEFT);
	
	GLUI_Rollout *p1 = glui->add_rollout("Generate", true);
	GLUI_Rollout *p2 = glui->add_rollout("Parameters", false);
	//GLUI_Rollout *p5 = glui->add_rollout("Hydraulic Weathering", false);
	GLUI_Rollout *p3 = glui->add_rollout("Display", true);
	GLUI_Rollout *p4 = glui->add_rollout("File System", false);
	glui->add_button("Quit", B_QUIT, ControlButton);
	
	//Generate
	glui->add_button_to_panel(p1, "Example", B_EXAMPLE, ControlButton);
	glui->add_button_to_panel(p1, "Random", B_RANDOM, ControlButton);
	glui->add_button_to_panel(p1, "Light", B_LIGHT, ControlButton);
	glui->add_button_to_panel(p1, "Wind", B_WIND, ControlButton);
	glui->add_separator_to_panel(p1);
	glui->add_button_to_panel(p1, "Step", B_STEP, ControlButton);
	glui->add_button_to_panel(p1, "Generate", B_GENERATE, ControlButton);
	glui->add_button_to_panel(p1, "Thickness", B_THICKNESS, ControlButton);
	
	//Parameters
	spin = glui->add_spinner_to_panel(p2, "Shadow Amt", GLUI_SPINNER_FLOAT, &shadowBase, B_PARAM, ControlButton);
	spin->set_float_limits(0.01, 10);
	spin = glui->add_spinner_to_panel(p2, "Shadow Spread", GLUI_SPINNER_FLOAT, &shadowFactor, B_PARAM, ControlButton);
	spin->set_float_limits(1, 10);
	spin = glui->add_spinner_to_panel(p2, "Max Light", GLUI_SPINNER_FLOAT, &lightMax, B_PARAM, ControlButton);
	spin->set_float_limits(1, 100);
	spin = glui->add_spinner_to_panel(p2, "Spread Depth", GLUI_SPINNER_INT, &qMax, B_PARAM, ControlButton);
	spin->set_float_limits(4, 100);
	spin = glui->add_spinner_to_panel(p2, "Wind Factor", GLUI_SPINNER_FLOAT, &WIND_FACTOR);
	spin->set_float_limits(1.0/16.0, 3);
	spin = glui->add_spinner_to_panel(p2, "Thickness N", GLUI_SPINNER_INT, &thicknessN, B_THICKNESS, ControlButton);
	spin->set_int_limits(2, 5);
	spin = glui->add_spinner_to_panel(p2, "Thickness Init", GLUI_SPINNER_INT, &thicknessInit, B_THICKNESS, ControlButton);
	spin->set_int_limits(1, 5);
	box = glui->add_listbox_to_panel(p2, "Apical Mode", (int*)&ltMode);
	box->add_item(DECURRENT, "Decurrent");
	box->add_item(NO_APICAL, "None");
	box->add_item(APICAL, "Apical");
	box->add_item(APICAL_MAIN, "Trunk");
	glui->add_checkbox_to_panel(p2, "Bend Branches", (int*)&windBendsBranches);
	
	//Display
	rotator = glui->add_rotation_to_panel(p3, "Rotation", rotation, B_ROTATION, ControlButton);
	
	//GLUI_Panel* p3_1 = glui->add_panel_to_panel(p3, "", GLUI_PANEL_NONE);
	glui->add_translation_to_panel(p3, "Move XY", GLUI_TRANSLATION_XY, &trans[0], B_TRANSLATE, ControlButton);
	glui->add_translation_to_panel(p3, "Move Z", GLUI_TRANSLATION_Z, &trans[2], B_TRANSLATE, ControlButton);
	glui->add_button_to_panel(p3, "+", B_ZOOM_IN, ControlButton);
	//glui->add_column_to_panel(p3_1, false);
	glui->add_button_to_panel(p3, "-", B_ZOOM_OUT, ControlButton);
	
	glui->add_button_to_panel(p3, "Reset", B_RESET_VIEW, ControlButton);
	
	box = glui->add_listbox_to_panel(p3, "Mode", (int*)&renderMode, B_MODE, ControlButton);
	box->add_item(MODE_STICK, "Stick");
	box->add_item(MODE_WIRE, "Wire");
	box->add_item(MODE_FILL, "Fill");
	box->add_item(MODE_SPHERE, "Spheres");
	
	//glui->add_checkbox_to_panel(p3, "Smooth Shading", (int*)&smoothShade, B_SHADING, ControlButton);
	//glui->add_checkbox_to_panel(p3, "Culling", (int*)&cull, B_CULL, ControlButton);
	glui->add_checkbox_to_panel(p3, "Generator Info", (int*)&showInfo, B_SHOW_INFO, ControlButton);
	glui->add_checkbox_to_panel(p3, "Inc. Grid", (int*)&showGrid, B_SHOW_INFO, ControlButton);
	
	//File System
	GLUI_EditText *txt = glui->add_edittext_to_panel(p4, "File Name:", GLUI_EDITTEXT_TEXT, filename);
	glui->add_button_to_panel(p4, "Load", B_LOAD, ControlButton);
	glui->add_button_to_panel(p4, "Save", B_SAVE, ControlButton);
	
	//Set default
	txt->set_text("tree.txt");
	
	glui->set_main_gfx_window(windowId);
	glui->sync_live();
}

int main(int argc, char** argv){
	
	glutInit(&argc, argv); 
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH); 
	glutInitWindowSize(WIDTH, HEIGHT); 
	glutInitWindowPosition(200, 100); 
	windowId = glutCreateWindow("Wind + Tree - Generater & Viewer");
	InitGUI(windowId);
	
	InitGL(); 
	
	//Texture loading
	//loadTextures();
	
	glutDisplayFunc(DrawScene); 
	GLUI_Master.set_glutIdleFunc(NULL);
	//GLUI_Master.set_glutKeyboardFunc(KeyDown);
	//GLUI_Master.set_glutSpecialFunc(SpecialKeyDown);
	//GLUI_Master.set_glutMouseFunc( myGlutMouse );
	//GLUI_Master.set_glutReshapeFunc( myGlutReshape );
	
	glutMainLoop();
	
	if( gen != NULL )
		delete gen;
	if( tree != NULL )
		delete tree;
	return 0;
}