#include "LightTree.h"
#include <stdlib.h>
#include <math.h>
#include <stack>
#include <algorithm>
#include "DrawVoxels.h"

#ifndef PI
#define PI 3.14159265
#endif


//Settings -- make non-const when want to have GUI change them
double W_MIN = 0.0006;
double W_MAX = 1.0;
double KAPPA = 0.5;	//0.35 makes taller tree -- weight drops to w_min at kappa*N
const double SEGMENT_LENGTH = 15;
const double ALPHA = 1.0;	//resource_base = alpha*light_base
const double THETA = PI/2;	//90 degrees perception volume cone
extern lighttree_mode_t ltMode;	//mode selected in GUI

//Function to calculate weight
#define W(i, n) ((i) < KAPPA*(n)? W_MAX + -(W_MAX - W_MIN)/(KAPPA*(n)) * (i) : W_MIN )


using namespace std;

/**
 * Generates a tree by stepping a few times.
 */
void LightTree::generate() {
	mode = ltMode;	//copy in GUI's setting
	
	//Do first bit with whatever apical setting
	while( iterations < 4 )
		step();
	
	//Then remove it later in development
	if( mode == APICAL || mode == APICAL_MAIN )
		mode = NO_APICAL;
	
	while( iterations < 6 )
		step();
	
	//Finally, draw thickness correctly
	clearThickness();
	calcThickness(thicknessInit);
}

/**
 * Draws debug info
 */
void LightTree::drawInfo() {
	draw_voxels(voxels, PIXELS_PER_VOXEL);
}


/**
 * Calculate the next season's growth based on environmental factors.
 * This involves:
 *	- Determine resources for each bud: gather info about light, use priority model to allcoation resource
 *	- Grow shoots. Buds left are those fated to grow.
 *		- In decurrent mode, terminating buds are removed, leaving only branching buds.
 *		- In apical control mode, terminating buds are placed on top of the priority list in resource allocation.
 *		- As a shoot grows, which floor(resource) segments with a bud at each one, buds cast shadows down.
 *	- Calculating the optimal shoot growth direction for a bud (for those that will grow next time)
 *		- observe shadow values in voxels within the bud's perception volume, which is conical with theta=90 deg and
 *		  r = 4-6 internode lengths.
 *		- pick the direction with least shadow.
 * 
 * Note: calculating bud vectors after growing shoots vs. before only means that root will grow up and not be
 * deflected arbitrarily (there's no shadow, so all neighbors are optimal).  Each bud will still be affected by
 * shadows from buds in the same and all previous iterations.
 */
void LightTree::step() {
	mode = ltMode;	//copy in GUI's setting
	
	printf("=== Begin Step %d ===\n", iterations);
	clearResource();
	gatherLight();
	giveResource();
	growShoots();
	calcVectors();
	printf("-== End Step %d ==-\n", iterations);
	
	iterations++;
	
	//This causes display list to be rebuilt, so it shows new shadowed regions
	decompile_voxels();
}

/**
 * Clear resource values before running the algorithm again.
 */
void LightTree::clearResource() {
	stack< Tree* > explore;
	
	
	explore.push(root);
	while( !explore.empty() ) {
		Tree * current = explore.top();
		explore.pop();
		printf("Clear (%5.3f,%5.3f,%5.3f)\n", current->x, current->y, current->z);
		
		current->light = 0;
		current->resource = 0;
		current->buds = 0;
		
		for(int i = 0; i < current->numChildren; i++) {
			explore.push(current->children[i]);
		}
	}
}


/**
 * Each bud collects light and passes information down the branches towards the base.
 */
void LightTree::gatherLight() {
	
	vector<Tree*>::const_iterator it;
	for(it=buds.begin(); it != buds.end(); it++) {
		Tree *current = *it;
		
		//Buds collect light and count it for this branch
		current->light = calcLight(current);
		current->buds = 1;
		printf("Gather at (%5.3f,%5.3f,%5.3f) = %5.3f\n", current->x, current->y, current->z, current->light);
		
		Tree *axisRoot = current->axisRoot;
		while( axisRoot != NULL ) {
			axisRoot->buds++;
			axisRoot->light += current->light;
			axisRoot = axisRoot->axisRoot;	//propagate all the way down to the base
		}
	}
	
	printf("Gathered light at base: %5.3f for %d buds\n", root->light, root->buds);
}

/**
 * Helper function for sorting lists of buds/branches by light/buds
 */
bool compare_branches(const Tree *a, const Tree *b) {
	//printf("~~~a.buds=%d, a.light=%5.3f, b.buds=%d, b.light=%5.3f\n", a->buds, a->light, b->buds, b->light);
	//return (a->buds != 0? (a->light / a->buds) : 0) < (b->buds != 0? (b->light / b->buds) : 0);
	return (a->buds != 0? (a->light / a->buds) : 0) > (b->buds != 0? (b->light / b->buds) : 0);
}

/**
 * Distributes resources through the tree to control how much each bud will grow.
 * 
 * Algorithm:
 *	go up the tree, for each axis root
 *		order buds and branches according to light/buds
 *		if apical control,
 *			put terminating bud on end of list -- do this by setting light to 9999
 *		else,
 *			in decurrent shouldn't be a terminating bud, so don't worry about it.
 *			in no_apical, it's already in the right place
 *		calculate v_i for each bud/branch according to forumula
 *		set the resource amount for that bud/branch
 */
void LightTree::giveResource() {
	stack< Tree* > explore;
	vector< Tree* > branches;
	vector<double> qw;
	
	//v_base = alpha * Q_base -- v is resource amount, Q is light
	root->resource = ALPHA * root->light;
	printf("Resource at base: %5.3f\n", root->resource);
	
	//Explore up the tree
	explore.push(root);
	while( !explore.empty() ) {
		Tree *current = explore.top();
		explore.pop();
		
		if( current->numChildren == 0 )
			continue;	//nothing to do
		
		printf("Allocating from axis (%5.3f,%5.3f,%5.3f)\n", current->x, current->y, current->z);
		//Scan through children.  If in same axis, ignore it but scan it's children.
		//If in a different axis, then it's a bud or branch.  Add to stack to continue
		//traversing the tree, also add to list for allocation at this stage.
		//A terminating bud is a bud that is the only child of a metamer.
		Tree *scan = current;
		Tree *term = NULL;		//will have to put light value back after sorting
		double termLight = 0;
		while( scan != NULL && scan->numChildren != 0 ) {
			const Tree *cur = scan;	//cache this, since we might change it before being done with it
			
			printf("Scanning from (%5.3f,%5.3f,%5.3f) -- (bud=%d,branch=%d) there are %d children\n", cur->x, cur->y, cur->z, cur->bud, cur->branch, cur->numChildren);
			for(int i = 0; i < cur->numChildren; i++) {
				Tree * child = cur->children[i];
				if( child->bud ) {
					//BTW, axisRoot would be the same, since buds point to parent's root for light gathering.
					if( cur->numChildren == 1 ) {
						//Terminating bud. (another check is if bud=1 and branch=0)
						printf("Found terminating bud (%5.3f,%5.3f,%5.3f)\n", child->x, child->y, child->z);
						if( mode == APICAL ) {
							termLight = child->light;
							term = child;
							child->light = 9999;	//this puts it on top of the priority list
						}
						branches.push_back(child);
						scan = child;	//will look here next and exit
					} else {
						//Normal bud.  Count it.
						printf("Found offshoot bud (%5.3f,%5.3f,%5.3f)\n", child->x, child->y, child->z);
						branches.push_back(child);
						
						if( child->branch == 0 ) {
							printf("WARNING: counted what is likely a terminal bud as offshoot bud\n");
						}
					}
				} else if( child->branch ) {
					//Marked as branch (will have same axis root, but be the axis root of others)
					printf("Found branch at (%5.3f,%5.3f,%5.3f)\n", child->x, child->y, child->z);
					branches.push_back(child);	//count it
					explore.push(child);		//and also continue allocating resources down it
					
					if( child->growth == cur->growth ) {
						printf("WARNING: counted what is likely a metamer as offshoot branch\n");
					}
					if( cur->numChildren == 1 ) {
						//in decurrent mode, can reach end with a branch vs. terminal bud
						scan = NULL;	//will end scan
					}
				} else {
					//Not a branch, so it's a metamer in the current branch (on the same axis)
					printf("Found metamer at (%5.3f,%5.3f,%5.3f)\n", child->x, child->y, child->z);
					scan = child;	//will look here next.  (Only ever one axis continuation)
				}
			}//check next child of cur
			
			if( scan == cur ) {
				printf("ERROR: didn't pick new scan node, will loop infinitely!\n");
				exit(-1);
			}
		}//continue scan up the axis
		
		//Order buds and branches by light/buds
		printf("Sorting %d buds and branches\n", (int)branches.size());
		sort(branches.begin(), branches.end(), compare_branches);
		
		if( term != NULL ) {
			term->light = termLight;	//put back this value now that sort is done
		}
		
		//Calc resource for each branch/bud
		vector<Tree*>::const_iterator it;
		int i, n = branches.size();
		double sumQW = 0;
		qw.reserve(n);	//will need this many spaces
		//Sum light * weight for each
		printf("Calculating Q_i * W_i and sum\n");
		for(i = 0,it = branches.begin(); it != branches.end(); it++, i++) {
			const Tree *cur = *it;
			qw[i] = W(i, n) * cur->light;
			sumQW += qw[i];
			printf("%d: @(%5.3f,%5.3f,%5.3f) W=%5.3f, Q=%5.3f, QW=%5.3f, Sum=%5.3f\n", i, cur->x, cur->y, cur->z, W(i,n), cur->light, qw[i], sumQW);
		}
		//Now use formula - sets resource amount
		printf("Passing out resource\n");
		if( sumQW == 0 ) {
			printf("WARNING: sumQW is zero -- this branch has no resource\n");
		} else {
			for(i = 0,it = branches.begin(); it != branches.end(); it++, i++) {
				Tree *cur = *it;
				cur->resource = current->resource * qw[i] / sumQW;
				printf("%d: @(%5.3f,%5.3f,%5.3f) rsrc=%5.3f\n", i, cur->x, cur->y, cur->z, cur->resource);
			}
		}
		
		//prepare for next pass
		qw.clear();
		branches.clear();
	}//and continue to next axis of the tree
}

/**
 * Send out shoots from buds with length based on allocated resources.
 * 
 * Algorithm:
 *	 for each bud
 *		#metamers = floor( resource )
 *		for 1 to #metamers
 *			add a bud and a segment (metamer) by:
 *				former terminal bud -> move out one segment length
 *				unset it's bud flag
 *				if not (decurrent and last metamer)
 *					add new terminal bud (loc is same as this new metamer's end loc)
 *				add new bud as child of metamer -- unless this is the last metamer (and #metamers > 1 or not decurrent)
 *			occludeDown from the new bud
 *			set new bud growth dir to be away from branch in some random direction
 */
void LightTree::growShoots() {
	vector<Tree*> newBuds;	//will hold all the new buds.  Old buds are all removed.
	
	vector<Tree*>::const_iterator it;
	for(it = buds.begin(); it != buds.end(); it++) {
		Tree *bud = *it;
		const int metamers = floor(bud->resource);
		
		printf("Adding %d metamers to shoot from (%5.3f,%5.3f,%5.3f) ->(%5.3f,%5.3f,%5.3f) (branch=%d)\n", metamers, bud->x, bud->y, bud->z, bud->growth.x, bud->growth.y, bud->growth.z, bud->branch);
		
		Tree *seg = bud;	//points to some bud to grow out
		for(int i = 0; i < metamers; i++) {
			//Vec3 loc = bud->growth * SEGMENT_LENGTH;
			//loc += Vec3(seg->x, seg->y, seg->z);
			//See if this has less rounding error??
			Vec3 loc = Vec3((double)bud->growth.x*SEGMENT_LENGTH + seg->x, (double)bud->growth.y*SEGMENT_LENGTH + seg->y, 
							(double)bud->growth.z*SEGMENT_LENGTH + seg->z);
			
			seg->bud = 0;
			printf("Adding one from (%5.3f,%5.3f,%5.3f) -> (%5.3f,%5.3f,%5.3f)\n", seg->x, seg->y, seg->z, loc.x, loc.y, loc.z);
			seg->x = loc.x;
			seg->y = loc.y;
			seg->z = loc.z;
			
			//Build children of seg
			Tree *term = NULL, *nbud = NULL;
			
			//no terminal on shoot's end for decurrent
			if( !(mode == DECURRENT && i == metamers-1) ) {
				term = new Tree(seg->x, seg->y, seg->z);
				term->thickness = 1;	//something visible
				term->bud = 1;
				term->growth = seg->growth;	//copy dir
				if( bud->branch )
					term->axisRoot = bud;	//Shoot grew from bud, so that's the root.
				else
					term->axisRoot = bud->axisRoot;	//bud was terminal bud, so in the same shoot as before
				
				if( i == metamers-1 )
					newBuds.push_back(term);	//we'll keep this terminal bud, so remember it
			}
			
			//if there's only one to add, or this is not the last segment (which doesn't get an off-shoot bud)
			if( (mode == DECURRENT && metamers == 1) || i != metamers-1 ) {
				nbud = new Tree(seg->x, seg->y, seg->z);
				nbud->thickness = 1;	//something visible
				nbud->bud = 1;
				nbud->branch = 1;	//different direction, will be root of a branch
				//Feeds light to parent shoot
				if( bud->branch )
					nbud->axisRoot = bud;	//Shoot grew from bud, so that's the root.
				else
					nbud->axisRoot = bud->axisRoot;	//bud was terminal bud, so in the same shoot as before
				
				//Get new growth dir -- I want something that is within 90 degrees of current direction, not
				//too close to current direction, and is random along the disk that sits in the orthogonal plane.
				//Cross-product will give me a vector in that disk.  Crossing with random vector will give it at
				//random place in the disk.  Linear interpolation with random factor between original and the
				//random vector in the disk gives me what I want.
				Vec3 inDisk;
				Vec3 rand;
				rand.StdRandomVector();
				printf("Random dir=(%5.3f,%5.3f,%5.3f)\n", rand.x, rand.y, rand.z);
				Vec3Cross(inDisk, rand, bud->growth);
				printf("Crossed inDisk dir=(%5.3f,%5.3f,%5.3f)\n", inDisk.x, inDisk.y, inDisk.z);
				const double s = drand48()*0.8 + 0.2;
				Vec3Lerp(nbud->growth, bud->growth, inDisk, s);	//don't be too similar to growth dir
				printf("Interpolated s=%5.3f dir=(%5.3f,%5.3f,%5.3f)\n", s, nbud->growth.x, nbud->growth.y, nbud->growth.z);
				nbud->growth.Normalize();
				
				//Cast a shadow
				occlude(nbud);
				
				newBuds.push_back(nbud);
			}
			
			if( term != NULL && nbud != NULL ) {
				seg->setNumChildren(2);
				seg->children[0] = term;
				seg->children[1] = nbud;
				printf("%d: Added bud @(%5.3f,%5.3f,%5.3f) ->(%5.3f,%5.3f,%5.3f) (bud=%d,branch=%d)\n", i, nbud->x, nbud->y, nbud->z, nbud->growth.x, nbud->growth.y, nbud->growth.z, nbud->bud, nbud->branch);
				printf("%d: Added terminal bud @(%5.3f,%5.3f,%5.3f) (bud=%d,branch=%d)\n", i, term->x, term->y, term->z, term->bud, term->branch);
			} else if( term != NULL ) {
				seg->setNumChildren(1);
				seg->children[0] = term;
				printf("%d: Added terminal bud @(%5.3f,%5.3f,%5.3f) (bud=%d,branch=%d)\n", i, term->x, term->y, term->z, term->bud, term->branch);
			} else if( nbud != NULL ) {
				seg->setNumChildren(1);
				seg->children[0] = nbud;
				printf("%d: Added bud @(%5.3f,%5.3f,%5.3f) ->(%5.3f,%5.3f,%5.3f) (bud=%d,branch=%d)\n", i, nbud->x, nbud->y, nbud->z, nbud->growth.x, nbud->growth.y, nbud->growth.z, nbud->bud, nbud->branch);
			}
			
			//Point to next terminal bud to expand into a metamer
			seg = term;	//only time this is null is when we're done (and it's in decurrent mode)
		}//continue with next metamer segment
	}//continue with next bud
	
	buds.swap(newBuds);	//constant time operation -- then we forget newBuds which holds the old list of buds
}

/**
 * Determines growth direction for all buds.
 * A perception volume is conical with theta=90 deg and r=4-6 internode lengths.
 * 
 * Algorithm:
 *	for each bud
 *		get neighboring voxels in perception volume
 *		choose one with least shadow
 *		set growth vector to point that way
 */
void LightTree::calcVectors() {
	const double r2 = 16 * SEGMENT_LENGTH * SEGMENT_LENGTH; //4 metamers distance
	const double theta_2 = THETA/2;	//can be +- this amount
	
	vector<Tree*>::const_iterator it;
	for(it = buds.begin(); it != buds.end(); it++) {
		Tree *bud = *it;
		
		if( bud->branch == 0 )
			continue;	//terminal bud, don't change growth direction
		
		//Get Neighbors
		Vec3i vox = voxels->pixelsToVoxel(bud->x, bud->y, bud->z, PIXELS_PER_VOXEL);
		Vec3i *neighbors = voxels->getNeighbors(vox.x, vox.y, vox.z);
		printf("Bud at (%5.3f,%5.3f,%5.3f) is in voxel (%d,%d,%d)\n", bud->x, bud->y, bud->z, vox.x, vox.y, vox.z);
		
		//Check them to find best direction
		Vec3i *least, *check;
		Vec3 leastDir;
		float shade = 999.0;
		least = NULL;
		
		
		check = &neighbors[0];
		while( check->x != -1 ) {
			//See if check is less shady than current least
			if( least != NULL && shade <= voxels->get(check->x, check->y, check->z) ) {
				check++;
				continue;
			}
			//See if check is in perception volume -- check center of voxel against perception cone
			const Vec3 center = voxels->voxelToPixels(check->x, check->y, check->z, PIXELS_PER_VOXEL);
			Vec3 centerDir = Vec3(center.x - bud->x, center.y - bud->y, center.z - bud->z);
			int dist2 = centerDir.LengthSq();
			if( dist2 > r2 ) {
				check++;
				continue;
			}
			centerDir.Normalize();
			double dot = Vec3Dot(centerDir, bud->growth);	//remember, growth should be normalized
			if(fabsf(dot) >= 1.0f) dot = 0.999f;	//copied from Vec3 -- probably accounts for float computation error?
			double angle = acos(dot);
			if( angle > PI ) angle -= PI;	//put in range -PI to PI
			if( fabs(angle) <= theta_2 ) {
				//Angle falls within perception cone, it's close enough, and it's less shady
				shade = voxels->get(check->x, check->y, check->z);
				least = check;
				leastDir = centerDir;
			}
			
			check++;
		}
		
		if( least != NULL ) {
			//Have a good direction to grow, so go that way.
			bud->growth = leastDir;
			printf("Least shady direction is (%5.3f,%5.3f,%5.3f)\n", leastDir.x, leastDir.y, leastDir.z);
		}
		
		//Free neighbors list
		free(neighbors);
	}//continue to next bud
}



/**
 * Find how much light exposure this bud has
 */
double LightTree::calcLight(Tree *bud) {
	//Find what voxel bud is in
	Vec3i v = voxels->pixelsToVoxel(bud->x, bud->y, bud->z, PIXELS_PER_VOXEL);
	//Get light at that voxel
	return voxels->exposure(v.x, v.y, v.z);
}

/**
 * Cast a shadow down.  This is a separate virtual function so WindTree can
 * also cast shelter sideways.
 */
void LightTree::occlude(Tree *bud) {
	Vec3i vox = voxels->pixelsToVoxel(bud->x, bud->y, bud->z, PIXELS_PER_VOXEL);
	voxels->occludeDown(vox);
}
