#pragma once

#include "Tree.h"
#include "Treegen.h"
#include "VoxelGrid.h"
#include <vector>


enum lighttree_mode_t {
	/* Produces bushes.  Done with proleptic development and sympodial branching. This means that new buds produce
	 * shoots in the next season (iteration) and that buds at the end of shoots die (or make flowers).
	 */
	DECURRENT,
	
	/* Produces trees that may be bush-like.  Shoots grow with a terminating bud, but no apical control, so they
	 * may grow out more rather than growing the branches longer.
	 */
	NO_APICAL,
	
	/* Produces trees.  This needs some kind of apical control to have lateral branches and tree top grow correctly.
	 * This means placing the terminating bud on the top of the priority list of resource allocation.
	 */
	APICAL,
	
	/*
	 * Produces tree with strong trunk and bushier branches/crown.  Apical control is only applied to the main
	 * trunk.  Note that you should remove apical control late in development as older trees become more decurrent.
	 */
	APICAL_MAIN
};


//Settings
#define PIXELS_PER_VOXEL 10
extern double W_MIN;
extern double W_MAX;
extern double KAPPA;


class LightTree : public Treegen
{
	protected:
	VoxelGrid *voxels;
	vector<Tree*> buds;
	
	public:
	lighttree_mode_t mode;
	
	/**
	 * Constructor
	 */
	LightTree()
	: voxels(NULL), mode(APICAL)
	{
		//Make a voxel grid to use
		voxels = new VoxelGrid(80,80,80);
		
		root->thickness = 1;	//so will be visible
		root->branch = 1;		//the main branch (trunk)
		
		//Make a first bud from root, direction is up
		root->setNumChildren(1);
		Tree *bud = new Tree(0,0,0);
		bud->bud = 1;
		bud->axisRoot = root;	//Light counted at bud and the bud's parent. It's in root's axis.
		bud->growth = root->growth = Vec3(0, 1, 0);
		bud->thickness = 1;	//so will be visible
		root->children[0] = bud;
		buds.push_back(bud);
		
		//Need to cast shadow for initial bud
		voxels->occludeDown(voxels->pixelsToVoxel(bud->x, bud->y, bud->z, PIXELS_PER_VOXEL));
	}
	
	/**
	 * Future work:
	 * To handle multiple self-organizing trees, pass the other one in here and then set the
	 * root of this tree to be something different.  This will remember the other tree 
	 * generator and will use the same voxel grid it has.  In the call to step, it will see
	 * there is another.  It will first do all calculations thru occlude but stop before 
	 * calc vectors.  Then it will call step on the other generator, then finish with calc
	 * vectors.  This allows stacking any number of generators and they will all put the
	 * shadows up first before any of them obsserves shadow.
	 */
	//LightTree(LightTree & other, float rootX, float rootY, float rootZ) ...
	
	
	virtual void generate();
	virtual void step();
	virtual void drawInfo();
	
	//Internal helper functions
	private:
	void clearResource();
	void gatherLight();
	void giveResource();
	void growShoots();
	
	double calcLight(Tree *bud);
	
	protected:
	virtual void calcVectors();	//virtual to allow a sub-class to add in wind effects.
	virtual void occlude(Tree *bud);
};