/*
Vector.h
Written by Matthew Fisher

The Vec3 struct represents a 3-dimensional (x, y, z) real ordered triplet.  It is used constantly everywhere.
The Vec4 struct represents a 4-dimensional (x, y, z, w) real ordered triplet.  It is used mostly for matrix multiplication.

Most of the Vec3 and Vec4 functions and overloaded operators are self-explanitory and are not commented.
*/

#ifndef __VEC_H
#define __VEC_H
#include <math.h>
#include <iostream>
#include <fstream>
using namespace std;

struct Vec3i {
	//Initalization
	Vec3i();
	Vec3i(const Vec3i &V);
	Vec3i(int _x, int _y, int _z);
	
	//Assignment
	Vec3i& operator = (const Vec3i &V);
	
	//Comparison
	 bool operator == (const Vec3i &V) const;
	 bool operator != (const Vec3i &V) const;	
	
	//Overloaded operators
	Vec3i& operator *= (int Right);
	Vec3i& operator /= (int Right);
	Vec3i& operator += (const Vec3i &Right);
	Vec3i& operator -= (const Vec3i &Right);
	
	int x, y, z;	//Local data
};


struct Vec3 {
	//Initalization
	 Vec3();
	 Vec3(const Vec3 &V);
	 Vec3(float _x, float _y, float _z);

	//Assignment
	 Vec3& operator = (const Vec3 &V);
	 void StdRandomVector();		//sets the current vector to be a point randomly chosen in the unit cube; (-1, 1) in all dimensions.
	 void StdRandomNormal();		//sets the current vector to be a point randomly chosen on the unit sphere.
	
	//Comparison
	 bool operator == (const Vec3 &V) const;
	 bool operator != (const Vec3 &V) const;	

	//Overloaded operators
	 Vec3& operator *= (float Right);
	 Vec3& operator /= (float Right);
	 Vec3& operator += (const Vec3 &Right);
	 Vec3& operator -= (const Vec3 &Right);

	//Normalization
	 void Normalize();			//normalizes the current vector
	 Vec3 ComponentSquare();

	//Accessors
	 float Length();
	 float LengthSq();			//returns the squared length of the vector

	// operator D3DXVECTOR3() const;

	float x, y, z;	//Local data
};

struct Vec4 {
	//Initalization
	 Vec4();
	 Vec4(const Vec4 &V);
	 Vec4(float _x, float _y, float _z, float _w);

	//Assignment
	 Vec4& operator = (const Vec4 &V);

	//Overloaded operators
	 Vec4& operator *= (float Right);
	 Vec4& operator /= (float Right);
	 Vec4& operator += (const Vec4 &Right);
	 Vec4& operator -= (const Vec4 &Right);
    bool operator != (const Vec4 &Right);

	//Normalization
	 void Normalize();

    void Homogenize();

	//Accessors
	 float Length();
	 float LengthSq();

	float x, y, z, w;	//Local data
};

//Functions associated with Vec3
 void Vec3TriNormal(Vec3 &Target, const Vec3 &V1, const Vec3 &V2, const Vec3 &V3);
 void Vec3Cross(Vec3 &Target, const Vec3 &Left, const Vec3 &Right);
 float Vec3Dot(const Vec3 &Left, const Vec3 &Right);
 void Vec3Lerp(Vec3 &Target, const Vec3 &Left, const Vec3 &Right, float s);	//linear interpolation between Left and Right by s; if s = 0, Left is returned, if s = 1, Right is returned.
 void Vec3Maximize(Vec3 &Target, const Vec3 &Left, const Vec3 &Right);		//returns the largest components in Left and Right in Max.
 void Vec3Minimize(Vec3 &Target, const Vec3 &Left, const Vec3 &Right);		//returns the smallest components in Left and Right in Max.

//Vec3 overloaded operators
 Vec3 operator * (const Vec3 &Left, float Right);
 Vec3 operator * (float Left, const Vec3 &Right);
 Vec3 operator / (const Vec3 &Left, float Right);
 Vec3 operator + (const Vec3 &Left, const Vec3 &Right);
 Vec3 operator - (const Vec3 &Left, const Vec3 &Right);
 Vec3 operator - (const Vec3 &V);


 float Vec4Dot(const Vec4 &Left, const Vec4 &Right);
 void Vec4Cross(Vec4 &Target, const Vec4 &Left, const Vec4 &Right);
 void Vec4Lerp(Vec4 &Target, const Vec4 &Left, const Vec4 &Right, float s);
 void Vec4Maximize(Vec4 &Target, const Vec4 &Left, const Vec4 &Right);
 void Vec4Minimize(Vec4 &Target, const Vec4 &Left, const Vec4 &Right);

//Vec4 overloaded operators
 Vec4 operator * (const Vec4 &Left, float Right);
 Vec4 operator * (float Left, const Vec4 &Right);
 Vec4 operator / (const Vec4 &Left, float Right);
 Vec4 operator + (const Vec4 &Left, const Vec4 &Right);
 Vec4 operator - (const Vec4 &Left, const Vec4 &Right);
 Vec4 operator - (const Vec4 &V);

 istream& operator >> (istream &os, Vec3 &v);				//inputs the vector's x, y, and z components
 ostream& operator << (ostream &os, const Vec3 &v);		//outputs the vector's x, y, and z components
 ostream& operator << (ostream &os, const Vec4 &v);		//outputs the vector's x, y, z, and w components

 float AngleBetween(const Vec3 &Left, const Vec3 &Right);	//returns the angle between the two vectors

#endif