#include "VoxelGrid.h"
#include <stdlib.h>

//Settings - used in casting shadow/shelter
float shadowBase = 1;		//a - amount of shadow cast by a bud
float shadowFactor = 2;		//b - divisor of shadow at each level down
float lightMax = 3;			//C - maximum light exposure in a cell
int qMax = 5;				//depth to decend in occluding

/**
 * Case shadow / shelter values through the grid. Shadow just needs
 * the location, and then goes down, but wind could be in an arbitrary
 * direction.
 * 
 * ...explain how...
 */
void VoxelGrid::occlude(Vec3i loc, Vec3 dir) {
	if( dir.x == 0 && dir.y == -1 && dir.z == 0 ) {
		//Going straight down, so use the quicker calculation
		occludeDown(loc);
	}
	if( dir.x == 1 && dir.y == 0 && dir.z == 0 ) {
		//Going straight right (x+), so use the quicker calculation
		occludeRight(loc);
	}
	
	//Need to cast the pyrimidal propogation in the given direction.
	
}

/**
 * Case shadow / shelter values through the grid, going straight down.
 * This is the quicker algorithm from Palubicki's paper.  A bud casts
 * a pyrimidal shadow straight down using (I +- p, J - q, K +- p)
 * where q = 0, 1, ..., qmax and p = 0, 1, ..., q.  The shadow in each
 * voxel is increased by delta-s = ab^-q where a > 0, b > 1 are given
 * constants.  Light exposure Q of a bud is Q = max(C − s + a, 0) where
 * C is a max (full) exposure constant. (The +a is because the bud 
 * doesn't shadow itself.)
 * 
 * Note: Shadows indicate optimal growth direction by considering all
 * neighboring voxels in the perception volume of the bud and selecting
 * the voxel with the lowest shadow.  Or else calculate the negative
 * gradient (which is the full calculus solution).
 */
void VoxelGrid::occludeDown(Vec3i loc) {
	printf("Doing occlude down...\n");
	//Do q=0 case first (since it's very simple and different when reduced)
	grid[I(loc.x, loc.y, loc.z)] += shadowBase;
	
	//Go down each level up to qMax
	for(int q = 1; q <= qMax; q++) {
		if( loc.y - q < 0 )
			break;	//reached the edge, so stop
		
		//Iterate through all combinations to fill out this level
		for(int p = -q; p <= q; p++) {
			for(int p2 = -q; p2 <= q; p2++) {
				if( loc.x + p < 0 || loc.x + p >= width 
				   || loc.z + p2 < 0 || loc.z + p2 >= depth )
					continue;	//out of bounds, so skip
				//printf("Updating loc (%d,%d,%d)\n", (int)loc.x + p, (int)loc.y - q, (int)loc.z + p2);
				grid[I(loc.x + p, loc.y - q, loc.z + p2)] += shadowBase * pow(shadowFactor, -q);
			}
		}
	}
	printf("finished occlude\n");
}

/**
 * Optimized like the occludeDown case, but going x+ direction (vs y-).
 * Uses (I + q, J +- p, K +- p), the rest is the same.
 */
void VoxelGrid::occludeRight(Vec3i loc) {
	//Do q=0 case first (since it's very simple and different when reduced)
	grid[I(loc.x, loc.y, loc.z)] += shadowBase;
	
	//Go down each level up to qMax
	for(int q = 1; q <= qMax; q++) {
		if( loc.x + q >= width )
			break;	//reached the edge, so stop
		
		//Iterate through all combinations to fill out this level
		for(int p = -q; p <= q; p++) {
			for(int p2 = -q; p2 <= q; p2++) {
				if( loc.y + p < 0 || loc.y + p >= width 
				   || loc.z + p2 < 0 || loc.z + p2 >= depth )
					continue;	//out of bounds, so skip
				
				grid[I(loc.x + q, loc.y + p, loc.z + p2)] += shadowBase * pow(shadowFactor, -q);
			}
		}
	}
}


/**
 * Light exposure Q of a bud is Q = max(C − s + a, 0) where
 * C is a max (full) exposure constant. (The +a is because the bud 
 * doesn't shadow itself.)
 */
float VoxelGrid::exposure(int x, int y, int z) {
	float expose = lightMax - get(x, y, z) + shadowBase;
	if( expose < 0 )
		return 0;
	else
		return expose;
}

/**
 * Convert from pixel coordinates to the corresponding voxel index.
 * Do bounds checking too (return edge boundary if off the edge).
 * 
 * This treats the bottom center of the voxel space as the origin for pixel space.
 * Also shifts by half a voxel in XZ so (0,0,0) is in middle of cell in middle of grid.
 */
Vec3i VoxelGrid::pixelsToVoxel(float px, float py, float pz, int pixPerVox) {
	int x, y, z;
	
	const float sx = width/2 * -pixPerVox - pixPerVox/2;
	const float sy = 0;
	const float sz = height/2 * -pixPerVox - pixPerVox/2;
	
	x = floor( (px - sx) / pixPerVox );
	y = floor( (py - sy) / pixPerVox );
	z = floor( (pz - sz) / pixPerVox );
	
	//Constrain to valid voxel
	if( x < 0 )
		x = 0;
	else if( x >= width )
		x = width - 1;
	
	if( y < 0 )
		y = 0;
	else if( y >= height )
		y = height - 1;
	
	if( z < 0 )
		z = 0;
	else if( z >= depth )
		z = depth - 1;
	
	return Vec3i(x, y, z);
}

/**
 * The reverse operation.  Doesn't need to do bounds checking.
 * @pre (x,y,z) is in bounds
 * @return location of center of the voxel
 */
Vec3 VoxelGrid::voxelToPixels(int vx, int vy, int vz, int pixPerVox) {
	float x, y, z;
	
	const float sx = width/2 * -pixPerVox - pixPerVox/2;
	const float sy = 0;
	const float sz = height/2 * -pixPerVox - pixPerVox/2;
	
	x = vx * pixPerVox + sx + pixPerVox/2;
	y = vy * pixPerVox + sy + pixPerVox/2;
	z = vz * pixPerVox + sz + pixPerVox/2;
	
	return Vec3(x, y, z);
}

/**
 * Get all voxels that are neigbors to the one specified.  This handles bounds checking too.  The
 * resultant array will have a (-1,-1,-1) element to signify the end. (And you should free it when 
 * done using it.)
 * 
 * @pre (x,y,z) is in bounds
 * @return (-1,-1,-1)-terminated array of all neighboring voxels (in bounds)
 */
Vec3i *VoxelGrid::getNeighbors(int x, int y, int z) {
	Vec3i *voxs;
	int c;
	
	//Get space for all + null.
	voxs = (Vec3i*)malloc(27 * sizeof(Vec3i));
	c = 0;
	
	if( x != 0 ) {
		//add voxels with x-1
		voxs[c++] = Vec3i(x-1, y, z);
		if( y != 0 ) {
			voxs[c++] = Vec3i(x-1, y-1, z);
			if( z != 0 )
				voxs[c++] = Vec3i(x-1, y-1, z-1);
			if( z != depth-1 )
				voxs[c++] = Vec3i(x-1, y-1, z+1);
		}
		if( y != height-1 ) {
			voxs[c++] = Vec3i(x-1, y+1, z);
			if( z != 0 )
				voxs[c++] = Vec3i(x-1, y+1, z-1);
			if( z != depth-1 )
				voxs[c++] = Vec3i(x-1, y+1, z+1);
		}
		if( z != 0 )
			voxs[c++] = Vec3i(x-1, y, z-1);
		if( z != depth-1 )
			voxs[c++] = Vec3i(x-1, y, z+1);
	}
	if( x != width - 1 ) {
		//add voxels with x+1
		voxs[c++] = Vec3i(x+1, y, z);
		if( y != 0 ) {
			voxs[c++] = Vec3i(x+1, y-1, z);
			if( z != 0 )
				voxs[c++] = Vec3i(x+1, y-1, z-1);
			if( z != depth-1 )
				voxs[c++] = Vec3i(x+1, y-1, z+1);
		}
		if( y != height-1 ) {
			voxs[c++] = Vec3i(x+1, y+1, z);
			if( z != 0 )
				voxs[c++] = Vec3i(x+1, y+1, z-1);
			if( z != depth-1 )
				voxs[c++] = Vec3i(x+1, y+1, z+1);
		}
		if( z != 0 )
			voxs[c++] = Vec3i(x+1, y, z-1);
		if( z != depth-1 )
			voxs[c++] = Vec3i(x+1, y, z+1);
	}
	//add voxels with x (but not x,y,z)
	if( y != 0 ) {
		voxs[c++] = Vec3i(x, y-1, z);
		if( z != 0 )
			voxs[c++] = Vec3i(x, y-1, z-1);
		if( z != depth-1 )
			voxs[c++] = Vec3i(x, y-1, z+1);
	}
	if( y != height-1 ) {
		voxs[c++] = Vec3i(x, y+1, z);
		if( z != 0 )
			voxs[c++] = Vec3i(x, y+1, z-1);
		if( z != depth-1 )
			voxs[c++] = Vec3i(x, y+1, z+1);
	}
	if( z != 0 )
		voxs[c++] = Vec3i(x, y, z-1);
	if( z != depth-1 )
		voxs[c++] = Vec3i(x, y, z+1);
	
	
	voxs[c] = Vec3i(-1,-1,-1);	//last unused spot is the terminator
	return voxs;
}
